# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description: 
"""

import os
import threading
import time
import wave
from pathlib import Path

import pyaudio
import pypinyin
from pydub import AudioSegment
from pypinyin import lazy_pinyin

from parrots.num_util import num2chinese
from parrots.utils.io_util import get_logger

default_logger = get_logger(__file__)


class TextToSpeech(object):
    def __init__(self, syllables_dir='./data/syllables'):
        self.syllables_dir = syllables_dir
        # TODO: 分数的读法 2.11 待修复，如何添加'.'
        self.punctuation = ['，', '。', '？', '！', '“', '”', '；', '：', '（', '）',
                            '.', ':', ';', ',', '?', '!', '\"', "\'", '(', ')']

    def _play_audio(self, path, delay, chunk=1024):
        try:
            time.sleep(delay)
            wf = wave.open(path, 'rb')
            p = pyaudio.PyAudio()
            stream = p.open(format=p.get_format_from_width(wf.getsampwidth()),
                            channels=wf.getnchannels(),
                            rate=wf.getframerate(),
                            output=True)

            data = wf.readframes(chunk)

            while data:
                stream.write(data)
                data = wf.readframes(chunk)

            stream.stop_stream()
            stream.close()

            p.terminate()
            return
        except IOError as ioe:
            default_logger.error(ioe)
        except Exception as e:
            default_logger.error(e)

    def speak(self, text):
        syllables = lazy_pinyin(text, style=pypinyin.TONE3)
        default_logger.info(syllables)
        delay = 0

        def preprocess(syllables):
            temp = []
            for syllable in syllables:
                for p in self.punctuation:
                    syllable = syllable.replace(p, '')
                if syllable.isdigit():
                    syllable = num2chinese(syllable)
                    new_sounds = lazy_pinyin(syllable, style=pypinyin.TONE3)
                    for e in new_sounds:
                        temp.append(e)
                else:
                    temp.append(syllable)
            return temp

        syllables = preprocess(syllables)
        threads = []
        for syllable in syllables:
            path = os.path.join(self.syllables_dir, syllable + ".wav")
            t = threading.Thread(target=self._play_audio, args=(path, delay))
            threads.append(t)
            delay += 0.355
        for t in threads:
            t.start()
        t.join()

    def synthesize(self, input_text='', output_wav_path='./out.wav'):
        """
        Synthesize .wav from text
        input_text: the folder that contains all syllables .wav files
        output_wav_path: the destination folder to save the synthesized file
        """
        delay = 0
        increment = 355  # milliseconds
        pause = 500  # pause for punctuation
        syllables = lazy_pinyin(input_text, style=pypinyin.TONE3)

        # initialize to be complete silence, each character takes up ~500ms
        result = AudioSegment.silent(duration=500 * len(input_text))
        for syllable in syllables:
            path = os.path.join(self.syllables_dir, syllable + ".wav")
            sound_file = Path(path)
            # insert 500 ms silence for punctuation marks
            if syllable in self.punctuation:
                short_silence = AudioSegment.silent(duration=pause)
                result = result.overlay(short_silence, position=delay)
                delay += increment
                continue
            # skip sound file that doesn't exist
            if not sound_file.is_file():
                continue
            segment = AudioSegment.from_wav(path)
            result = result.overlay(segment, position=delay)
            delay += increment

        result.export(output_wav_path, format="wav")
        print(output_wav_path)
        default_logger.info("Exported:" + output_wav_path)



# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description: 
"""
import os

from .pinyin2hanzi import Pinyin2Hanzi
from .speech_recognition import SpeechRecognition
from .tts import TextToSpeech

__version__ = '0.1.1'
_pwd_path = os.path.abspath(os.path.dirname(__file__))
_get_abs_path = lambda path: os.path.normpath(os.path.join(_pwd_path, path))

DEFAULT_PINYIN_DICT = _get_abs_path('data/pinyin_hanzi_dict.txt')
DEFAULT_SPEECH_MODEL = _get_abs_path('data/speech_model/speech_recognition.model')
DEFAULT_PINYIN2HANZI_DIR = _get_abs_path('data/pinyin2hanzi')
DEFAULT_SYLLABLES_DIR = _get_abs_path('data/syllables')

ms = SpeechRecognition(pinyin_path=DEFAULT_PINYIN_DICT,
                       model_path=DEFAULT_SPEECH_MODEL)
recognize_speech_from_file = ms.recognize_speech_from_file

p2h = Pinyin2Hanzi(model_dir=DEFAULT_PINYIN2HANZI_DIR)
pinyin_2_hanzi = p2h.pinyin_2_hanzi

t2s = TextToSpeech(syllables_dir=DEFAULT_SYLLABLES_DIR)
speak = t2s.speak
synthesize = t2s.synthesize



# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description: 获取符号字典列表的程序
"""


def get_pinyin_list(dict_path=''):
    """
    加载拼音符号列表，用于标记符号
    :param dict_path: 拼音符号列表
    :return:
    """
    list_symbol = []  # 符号列表
    pinyin_idx = 0
    with open(dict_path, mode='r', encoding='UTF-8') as f:
        for line in f:
            line = line.strip('\n')
            parts = line.split('\t')
            list_symbol.append(parts[pinyin_idx])
    list_symbol.append('_')
    return list_symbol



# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description:
"""

import json
import sys
from pathlib import Path

from pydub import AudioSegment
from pydub.silence import split_on_silence

with open('mapping.json') as data_file:
    data = json.load(data_file)

# keys = sorted(data.keys())
if len(sys.argv) != 2:
    print("usage:")
    print("      python process.py {key}")
    print("to process ./recording/{key}.wav")
    sys.exit(1)

key = sys.argv[1:][0]

syllables = data[key]

path = "./recording/" + key + ".wav"

file = Path(path)
if not file.is_file():
    raise Exception(path + " doesn't exist")

sound_file = AudioSegment.from_wav(path)

audio_chunks = split_on_silence(sound_file,
                                # must be silent for at least 300ms
                                min_silence_len=300,
                                # consider it silent if quieter than -48 dBFS
                                silence_thresh=-48
                                )

flag = False
if len(syllables) * 5 != len(audio_chunks):
    flag = True

for i, chunk in enumerate(audio_chunks):
    syllable = syllables[i // 5]
    print(syllable)
    j = i % 5
    if j != 4:  # 1st, 2nd, 3rd, 4th tone
        out_file = "./pre/" + syllable + str(j + 1) + ".wav"
    else:  # neutrual tone
        out_file = "./pre/" + syllable + ".wav"

    # out_file = "./pre/"+ str(i)+".wav"
    print("exporting", out_file)
    chunk.export(out_file, format="wav")

print(key, len(audio_chunks))
if flag:
    print(key, "doesn't match expected audio chunks length")
    print("expected:", len(syllables) * 5, "actual:", len(audio_chunks))



# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description: math
"""

import difflib


def edit_distance(str1, str2):
    distance = 0
    s = difflib.SequenceMatcher(None, str1, str2)
    for tag, i1, i2, j1, j2 in s.get_opcodes():
        # print('{:7} a[{}: {}] --> b[{}: {}] {} --> {}'.format(tag, i1, i2, j1, j2, str1[i1: i2], str2[j1: j2]))
        if tag == 'replace':
            distance += max(i2 - i1, j2 - j1)
        elif tag == 'insert':
            distance += (j2 - j1)
        elif tag == 'delete':
            distance += (i2 - i1)
    return distance



# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description:
"""
import math
import wave

import matplotlib.pyplot as plt
import numpy as np
from python_speech_features import delta
from python_speech_features import mfcc
from scipy.fftpack import fft


def read_wav_data(filename):
    '''
    读取一个wav文件，返回声音信号的时域谱矩阵和播放时间
    '''
    wav = wave.open(filename, "rb")  # 打开一个wav格式的声音文件流
    num_frame = wav.getnframes()  # 获取帧数
    num_channel = wav.getnchannels()  # 获取声道数
    framerate = wav.getframerate()  # 获取帧速率
    str_data = wav.readframes(num_frame)  # 读取全部的帧
    wav.close()  # 关闭流
    wave_data = np.fromstring(str_data, dtype=np.short)  # 将声音文件数据转换为数组矩阵形式
    wave_data.shape = -1, num_channel  # 按照声道数将数组整形，单声道时候是一列数组，双声道时候是两列的矩阵
    wave_data = wave_data.T  # 将矩阵转置
    return wave_data, framerate


def GetMfccFeature(wavsignal, fs):
    # 获取输入特征
    feat_mfcc = mfcc(wavsignal[0], fs)
    feat_mfcc_d = delta(feat_mfcc, 2)
    feat_mfcc_dd = delta(feat_mfcc_d, 2)
    # 返回值分别是mfcc特征向量的矩阵及其一阶差分和二阶差分矩阵
    wav_feature = np.column_stack((feat_mfcc, feat_mfcc_d, feat_mfcc_dd))
    return wav_feature


def GetFrequencyFeature(wavsignal, fs):
    # wav波形 加时间窗以及时移10ms
    time_window = 25  # 单位ms
    data_input = []

    # print(int(len(wavsignal[0])/fs*1000 - time_window) // 10)
    wav_length = len(wavsignal[0])  # 计算一条语音信号的原始长度
    range0_end = int(len(wavsignal[0]) / fs * 1000 - time_window) // 10  # 计算循环终止的位置，也就是最终生成的窗数
    for i in range(0, range0_end):
        p_start = i * 160
        p_end = p_start + 400
        data_line = []

        for j in range(p_start, p_end):
            data_line.append(wavsignal[0][j])
        # print('wavsignal[0][j]:\n',wavsignal[0][j])
        # data_line = abs(fft(data_line)) / len(wavsignal[0])
        data_line = fft(data_line) / wav_length
        data_line2 = []
        for fre_sig in data_line:
            # 分别取出频率信号的实部和虚部作为语音信号的频率特征
            # 直接使用复数的话，之后会被numpy将虚部丢弃，造成信息丢失
            # print('fre_sig:\n',fre_sig)
            data_line2.append(fre_sig.real)
            data_line2.append(fre_sig.imag)

        data_input.append(data_line2[0:len(data_line2) // 2])  # 除以2是取一半数据，因为是对称的
    return data_input


def GetFrequencyFeature2(wavsignal, fs):
    # wav波形 加时间窗以及时移10ms
    time_window = 25  # 单位ms
    window_length = fs / 1000 * time_window  # 计算窗长度的公式，目前全部为400固定值

    wav_arr = np.array(wavsignal)
    # wav_length = len(wavsignal[0])
    wav_length = wav_arr.shape[1]

    range0_end = int(len(wavsignal[0]) / fs * 1000 - time_window) // 10  # 计算循环终止的位置，也就是最终生成的窗数
    data_input = np.zeros((range0_end, 200), dtype=np.float)  # 用于存放最终的频率特征数据
    data_line = np.zeros((1, 400), dtype=np.float)
    for i in range(0, range0_end):
        p_start = i * 160
        p_end = p_start + 400

        data_line = wav_arr[0, p_start:p_end]
        '''
        x=np.linspace(0, 400 - 1, 400, dtype = np.int64)
        w = 0.54 - 0.46 * np.cos(2 * np.pi * (x) / (400 - 1) ) # 汉明窗
        data_line = data_line * w # 加窗
        '''
        data_line = np.abs(fft(data_line)) / wav_length

        data_input[i] = data_line[0:200]  # 设置为400除以2的值（即200）是取一半数据，因为是对称的

    # print(data_input.shape)
    return data_input


x = np.linspace(0, 400 - 1, 400, dtype=np.int64)
w = 0.54 - 0.46 * np.cos(2 * np.pi * (x) / (400 - 1))  # 汉明窗


def get_frequency_features(wavsignal, fs):
    # wav波形 加时间窗以及时移10ms
    time_window = 25  # 单位ms
    window_length = fs / 1000 * time_window  # 计算窗长度的公式，目前全部为400固定值

    wav_arr = np.array(wavsignal)
    wav_length = wav_arr.shape[1]

    range0_end = int(len(wavsignal[0]) / fs * 1000 - time_window) // 10  # 计算循环终止的位置，也就是最终生成的窗数
    data_input = np.zeros((range0_end, 200), dtype=np.float)  # 用于存放最终的频率特征数据
    data_line = np.zeros((1, 400), dtype=np.float)

    for i in range(0, range0_end):
        p_start = i * 160
        p_end = p_start + 400
        data_line = wav_arr[0, p_start:p_end]
        data_line = data_line * w  # 加窗
        data_line = np.abs(fft(data_line)) / wav_length
        data_input[i] = data_line[0:200]  # 设置为400除以2的值（即200）是取一半数据，因为是对称的

    # print(data_input.shape)
    data_input = np.log(data_input + 1)
    return data_input


def wav_scale(energy):
    '''
    语音信号能量归一化
    '''
    means = energy.mean()  # 均值
    var = energy.var()  # 方差
    e = (energy - means) / math.sqrt(var)  # 归一化能量
    return e


def wav_scale2(energy):
    '''
    语音信号能量归一化
    '''
    maxnum = max(energy)
    e = energy / maxnum
    return e


def wav_scale3(energy):
    '''
    语音信号能量归一化
    '''
    for i in range(len(energy)):
        # if i == 1:
        #	#print('wavsignal[0]:\n {:.4f}'.format(energy[1]),energy[1] is int)
        energy[i] = float(energy[i]) / 100.0
    # if i == 1:
    #	#print('wavsignal[0]:\n {:.4f}'.format(energy[1]),energy[1] is int)
    return energy


def wav_show(wave_data, fs):  # 显示出来声音波形
    time = np.arange(0, len(wave_data)) * (1.0 / fs)  # 计算声音的播放时间，单位为秒
    # 画声音波形
    # plt.subplot(211)
    plt.plot(time, wave_data)
    # plt.subplot(212)
    # plt.plot(time, wave_data[1], c = "g")
    plt.show()


def get_wav_list(filename):
    '''
    读取一个wav文件列表，返回一个存储该列表的字典类型值
    ps:在数据中专门有几个文件用于存放用于训练、验证和测试的wav文件列表
    '''
    txt_obj = open(filename, 'r')  # 打开文件并读入
    txt_text = txt_obj.read()
    txt_lines = txt_text.split('\n')  # 文本分割
    dic_filelist = {}  # 初始化字典
    list_wavmark = []  # 初始化wav列表
    for i in txt_lines:
        if (i != ''):
            txt_l = i.split(' ')
            dic_filelist[txt_l[0]] = txt_l[1]
            list_wavmark.append(txt_l[0])
    txt_obj.close()
    return dic_filelist, list_wavmark


def get_wav_symbol(filename):
    '''
    读取指定数据集中，所有wav文件对应的语音符号
    返回一个存储符号集的字典类型值
    '''
    txt_obj = open(filename, 'r')  # 打开文件并读入
    txt_text = txt_obj.read()
    txt_lines = txt_text.split('\n')  # 文本分割
    dic_symbol_list = {}  # 初始化字典
    list_symbolmark = []  # 初始化symbol列表
    for i in txt_lines:
        if (i != ''):
            txt_l = i.split(' ')
            dic_symbol_list[txt_l[0]] = txt_l[1:]
            list_symbolmark.append(txt_l[0])
    txt_obj.close()
    return dic_symbol_list, list_symbolmark



# -*- coding: utf-8 -*-
# Author: XuMing <xuming624@qq.com>
# Brief: 
import logging
import os
import pickle


def get_logger(name, log_file=None):
    """
    logger
    :param name: 模块名称
    :param log_file: 日志文件，如无则输出到标准输出
    :return:
    """
    format = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    if not log_file:
        handle = logging.StreamHandler()
    else:
        handle = logging.FileHandler(log_file)
    handle.setFormatter(format)
    logger = logging.getLogger(name)
    logger.addHandler(handle)
    logger.setLevel(logging.DEBUG)
    return logger


def load_pkl(pkl_path):
    """
    加载词典文件
    :param pkl_path:
    :return:
    """
    with open(pkl_path, 'rb') as f:
        result = pickle.load(f)
    return result


def dump_pkl(vocab, pkl_path, overwrite=True):
    """
    存储文件
    :param pkl_path:
    :param overwrite:
    :return:
    """
    if os.path.exists(pkl_path) and not overwrite:
        return
    with open(pkl_path, 'wb') as f:
        # pickle.dump(vocab, f, protocol=pickle.HIGHEST_PROTOCOL)
        pickle.dump(vocab, f, protocol=0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: nl8590687
语音识别的语言模型

基于马尔可夫模型的语言模型
"""
import os


class Pinyin2Hanzi(object):
    def __init__(self, model_dir='./data/pinyin2hanzi'):
        self.dict_pinyin = self.get_symbol_dict(os.path.join(model_dir, '..', 'pinyin_hanzi_dict.txt'))
        self.model1 = self.get_model_file(os.path.join(model_dir, 'char_idx.txt'))
        self.model2 = self.get_model_file(os.path.join(model_dir, 'word_idx.txt'))
        self.pinyin = self.get_pinyin(os.path.join(model_dir, 'dic_pinyin.txt'))
        self.model = (self.dict_pinyin, self.model1, self.model2)

    def pinyin_2_hanzi(self, list_syllable):
        """
        语音拼音 => 文本
        :param list_syllable:
        :return:
        """
        r = ''
        length = len(list_syllable)
        if not length:  # 传入的参数没有包含任何拼音时
            return ''

        # 先取出一个字，即拼音列表中第一个字
        str_tmp = [list_syllable[0]]

        for i in range(0, length - 1):
            # 依次从第一个字开始每次连续取两个字拼音
            str_split = list_syllable[i] + ' ' + list_syllable[i + 1]
            # print(str_split,str_tmp,r)
            # 如果这个拼音在汉语拼音状态转移字典里的话
            if (str_split in self.pinyin):
                # 将第二个字的拼音加入
                str_tmp.append(list_syllable[i + 1])
            else:
                # 否则不加入，然后直接将现有的拼音序列进行解码
                str_decode = self.decode(str_tmp, 0.0000)
                # print('decode ',str_tmp,str_decode)
                if (str_decode != []):
                    r += str_decode[0][0]
                # 再重新从i+1开始作为第一个拼音
                str_tmp = [list_syllable[i + 1]]
        # print('最后：', str_tmp)
        str_decode = self.decode(str_tmp, 0.0000)
        if str_decode:
            r += str_decode[0][0]
        return r

    def decode(self, list_syllable, yuzhi=0.0001):
        '''
        实现拼音向文本的转换
        基于马尔可夫链
        '''
        # assert self.dic_pinyin == null or self.model1 == null or self.model2 == null
        list_words = []

        num_pinyin = len(list_syllable)
        # print('======')
        # print('decode function: list_syllable\n',list_syllable)
        # print(num_pinyin)
        # 开始语音解码
        for i in range(num_pinyin):
            # print(i)
            ls = ''
            if (list_syllable[i] in self.dict_pinyin):  # 如果这个拼音在汉语拼音字典里的话
                # 获取拼音下属的字的列表，ls包含了该拼音对应的所有的字
                ls = self.dict_pinyin[list_syllable[i]]
            else:
                break

            if (i == 0):
                # 第一个字做初始处理
                num_ls = len(ls)
                for j in range(num_ls):
                    tuple_word = ['', 0.0]
                    # 设置马尔科夫模型初始状态值
                    # 设置初始概率，置为1.0
                    tuple_word = [ls[j], 1.0]
                    # print(tuple_word)
                    # 添加到可能的句子列表
                    list_words.append(tuple_word)

                # print(list_words)
                continue
            else:
                # 开始处理紧跟在第一个字后面的字
                list_words_2 = []
                num_ls_word = len(list_words)
                # print('ls_wd: ',list_words)
                for j in range(0, num_ls_word):

                    num_ls = len(ls)
                    for k in range(0, num_ls):
                        tuple_word = ['', 0.0]
                        tuple_word = list(list_words[j])  # 把现有的每一条短语取出来
                        # print('tw1: ',tuple_word)
                        tuple_word[0] = tuple_word[0] + ls[k]  # 尝试按照下一个音可能对应的全部的字进行组合
                        # print('ls[k]  ',ls[k])

                        tmp_words = tuple_word[0][-2:]  # 取出用于计算的最后两个字
                        # print('tmp_words: ',tmp_words,tmp_words in self.model2)
                        if (tmp_words in self.model2):  # 判断它们是不是再状态转移表里
                            # print(tmp_words,tmp_words in self.model2)
                            tuple_word[1] = tuple_word[1] * float(self.model2[tmp_words]) / float(
                                self.model1[tmp_words[-2]])
                        # 核心！在当前概率上乘转移概率，公式化简后为第n-1和n个字出现的次数除以第n-1个字出现的次数
                        # print(self.model2[tmp_words],self.model1[tmp_words[-2]])
                        else:
                            tuple_word[1] = 0.0
                            continue
                        # print('tw2: ',tuple_word)
                        # print(tuple_word[1] >= pow(yuzhi, i))
                        if (tuple_word[1] >= pow(yuzhi, i)):
                            # 大于阈值之后保留，否则丢弃
                            list_words_2.append(tuple_word)

                list_words = list_words_2
                # print(list_words,'\n')
        # print(list_words)
        for i in range(0, len(list_words)):
            for j in range(i + 1, len(list_words)):
                if (list_words[i][1] < list_words[j][1]):
                    tmp = list_words[i]
                    list_words[i] = list_words[j]
                    list_words[j] = tmp

        return list_words

    def get_symbol_dict(self, file_path):
        """
        读取拼音汉字的字典文件
        :param file_path:
        :return: 读取后的字典
        """
        txt_obj = open(file_path, 'r', encoding='UTF-8')  # 打开文件并读入
        txt_text = txt_obj.read()
        txt_obj.close()
        txt_lines = txt_text.split('\n')  # 文本分割

        dic_symbol = {}  # 初始化符号字典
        for i in txt_lines:
            list_symbol = []  # 初始化符号列表
            if i:
                txt_l = i.split('\t')
                pinyin = txt_l[0]
                for word in txt_l[1]:
                    list_symbol.append(word)
            dic_symbol[pinyin] = list_symbol
        return dic_symbol

    def get_model_file(self, model_path):
        """
        读取语言模型的文件
        :param model_path:
        :return: 读取后的模型
        """
        txt_obj = open(model_path, 'r', encoding='UTF-8')  # 打开文件并读入
        txt_text = txt_obj.read()
        txt_obj.close()
        txt_lines = txt_text.split('\n')  # 文本分割

        dic_model = {}  # 初始化符号字典
        for i in txt_lines:
            if i:
                txt_l = i.split('\t')
                if (len(txt_l) == 1):
                    continue
                dic_model[txt_l[0]] = txt_l[1]
        return dic_model

    def get_pinyin(self, filename):
        file_obj = open(filename, 'r', encoding='UTF-8')
        txt_all = file_obj.read()
        file_obj.close()

        txt_lines = txt_all.split('\n')
        dic = {}
        for line in txt_lines:
            if not line:
                continue
            pinyin_split = line.split('\t')
            list_pinyin = pinyin_split[0]
            if (list_pinyin not in dic) and int(pinyin_split[1]) > 1:
                dic[list_pinyin] = pinyin_split[1]
        return dic



# -*- coding: utf-8 -*-
"""
@author:XuMing（xuming624@qq.com)
@description:
"""
import os
import numpy as np
from keras import backend as K
from keras.layers import Dense, Dropout, Input, Reshape  # , Flatten
from keras.layers import Lambda, Activation, Conv2D, MaxPooling2D  # , Merge
from keras.models import Model
from keras.optimizers import Adam

from parrots.utils.file_reader import get_pinyin_list
from parrots.utils.wav_util import get_frequency_features, read_wav_data
from parrots.utils.io_util import get_logger
import time

default_logger = get_logger(__file__)


class SpeechRecognition(object):
    def __init__(self, pinyin_path='./data/pinyin_hanzi_dict.txt',
                 model_path='./data/speech_model/speech_recognition.model'):
        """
        初始化
        :param pinyin_path: pinyin file path
        """
        MS_OUTPUT_SIZE = 1422  # 默认输出的拼音的表示大小是1422，即1421个拼音+1个空白块
        self.MS_OUTPUT_SIZE = MS_OUTPUT_SIZE  # 神经网络最终输出的每一个字符向量维度的大小
        self.label_max_string_length = 64
        self.AUDIO_LENGTH = 1600
        self.AUDIO_FEATURE_LENGTH = 200
        self._model, self.base_model = self.create_model()
        self.initialized = False
        self.pinyin_path = pinyin_path
        self.model_path = model_path

    def initialize(self):
        pwd_path = os.path.abspath(os.path.dirname(__file__))
        if self.pinyin_path:
            t1 = time.time()
            try:
                self.pinyin_list = get_pinyin_list(self.pinyin_path)  # 获取拼音列表
            except IOError:
                pinyin_path = os.path.join(pwd_path, '..', self.pinyin_path)
                self.pinyin_list = get_pinyin_list(pinyin_path)  # 获取拼音列表
            default_logger.debug(
                "Loading pinyin dict cost %.3f seconds." % (time.time() - t1))
        if self.model_path:
            t2 = time.time()
            try:
                self._model.load_weights(self.model_path)
                self.base_model.load_weights(self.model_path + '.base')
            except IOError:
                model_path = os.path.join(pwd_path, '..', self.model_path)
                self._model.load_weights(model_path)
                self.base_model.load_weights(model_path + '.base')
            self.initialized = True
            default_logger.debug(
                "Loading model cost %.3f seconds." % (time.time() - t2))
            default_logger.debug("Speech recognition model has been built ok.")

    def check_initialized(self):
        if not self.initialized:
            self.initialize()

    def create_model(self):
        """
        定义CNN/LSTM/CTC模型，使用函数式模型
        输入层：200维的特征值序列，一条语音数据的最大长度设为1600（大约16s）
        隐藏层：卷积池化层，卷积核大小为3x3，池化窗口大小为2
        隐藏层：全连接层
        输出层：全连接层，神经元数量为self.MS_OUTPUT_SIZE，使用softmax作为激活函数，
        CTC层：使用CTC的loss作为损失函数，实现连接性时序多输出
        :return:
        """
        input_data = Input(name='the_input', shape=(self.AUDIO_LENGTH, self.AUDIO_FEATURE_LENGTH, 1))

        layer_h1 = Conv2D(32, (3, 3), use_bias=False, activation='relu', padding='same',
                          kernel_initializer='he_normal')(input_data)  # 卷积层
        layer_h1 = Dropout(0.05)(layer_h1)
        layer_h2 = Conv2D(32, (3, 3), use_bias=True, activation='relu', padding='same', kernel_initializer='he_normal')(
            layer_h1)  # 卷积层
        layer_h3 = MaxPooling2D(pool_size=2, strides=None, padding="valid")(layer_h2)  # 池化层
        # layer_h3 = Dropout(0.2)(layer_h2) # 随机中断部分神经网络连接，防止过拟合
        layer_h3 = Dropout(0.05)(layer_h3)
        layer_h4 = Conv2D(64, (3, 3), use_bias=True, activation='relu', padding='same', kernel_initializer='he_normal')(
            layer_h3)  # 卷积层
        layer_h4 = Dropout(0.1)(layer_h4)
        layer_h5 = Conv2D(64, (3, 3), use_bias=True, activation='relu', padding='same', kernel_initializer='he_normal')(
            layer_h4)  # 卷积层
        layer_h6 = MaxPooling2D(pool_size=2, strides=None, padding="valid")(layer_h5)  # 池化层

        layer_h6 = Dropout(0.1)(layer_h6)
        layer_h7 = Conv2D(128, (3, 3), use_bias=True, activation='relu', padding='same',
                          kernel_initializer='he_normal')(layer_h6)  # 卷积层
        layer_h7 = Dropout(0.15)(layer_h7)
        layer_h8 = Conv2D(128, (3, 3), use_bias=True, activation='relu', padding='same',
                          kernel_initializer='he_normal')(layer_h7)  # 卷积层
        layer_h9 = MaxPooling2D(pool_size=2, strides=None, padding="valid")(layer_h8)  # 池化层

        layer_h9 = Dropout(0.15)(layer_h9)
        layer_h10 = Conv2D(128, (3, 3), use_bias=True, activation='relu', padding='same',
                           kernel_initializer='he_normal')(layer_h9)  # 卷积层
        layer_h10 = Dropout(0.2)(layer_h10)
        layer_h11 = Conv2D(128, (3, 3), use_bias=True, activation='relu', padding='same',
                           kernel_initializer='he_normal')(layer_h10)  # 卷积层
        layer_h12 = MaxPooling2D(pool_size=1, strides=None, padding="valid")(layer_h11)  # 池化层

        layer_h12 = Dropout(0.2)(layer_h12)
        layer_h13 = Conv2D(128, (3, 3), use_bias=True, activation='relu', padding='same',
                           kernel_initializer='he_normal')(layer_h12)  # 卷积层
        layer_h13 = Dropout(0.2)(layer_h13)
        layer_h14 = Conv2D(128, (3, 3), use_bias=True, activation='relu', padding='same',
                           kernel_initializer='he_normal')(layer_h13)  # 卷积层
        layer_h15 = MaxPooling2D(pool_size=1, strides=None, padding="valid")(layer_h14)  # 池化层

        # test=Model(inputs = input_data, outputs = layer_h12)
        # test.summary()

        layer_h16 = Reshape((200, 3200))(layer_h15)  # Reshape层
        # layer_h5 = LSTM(256, activation='relu', use_bias=True, return_sequences=True)(layer_h4) # LSTM层
        # layer_h6 = Dropout(0.2)(layer_h5) # 随机中断部分神经网络连接，防止过拟合
        layer_h16 = Dropout(0.3)(layer_h16)
        layer_h17 = Dense(128, activation="relu", use_bias=True, kernel_initializer='he_normal')(layer_h16)  # 全连接层
        layer_h17 = Dropout(0.3)(layer_h17)
        layer_h18 = Dense(self.MS_OUTPUT_SIZE, use_bias=True, kernel_initializer='he_normal')(layer_h17)  # 全连接层

        y_pred = Activation('softmax', name='Activation0')(layer_h18)
        model_data = Model(inputs=input_data, outputs=y_pred)
        # model_data.summary()

        labels = Input(name='the_labels', shape=[self.label_max_string_length], dtype='float32')
        input_length = Input(name='input_length', shape=[1], dtype='int64')
        label_length = Input(name='label_length', shape=[1], dtype='int64')
        # Keras doesn't currently support loss funcs with extra parameters
        # so CTC loss is implemented in a lambda layer

        # layer_out = Lambda(ctc_lambda_func,output_shape=(self.MS_OUTPUT_SIZE, ), name='ctc')([y_pred, labels, input_length, label_length])#(layer_h6) # CTC
        loss_out = Lambda(self.ctc_lambda_func, output_shape=(1,), name='ctc')(
            [y_pred, labels, input_length, label_length])

        model = Model(inputs=[input_data, labels, input_length, label_length], outputs=loss_out)

        # model.summary()

        # clipnorm seems to speeds up convergence
        # sgd = SGD(lr=0.0001, decay=1e-6, momentum=0.9, nesterov=True, clipnorm=5)
        # opt = Adadelta(lr = 0.01, rho = 0.95, epsilon = 1e-06)
        opt = Adam(lr=0.001, beta_1=0.9, beta_2=0.999, decay=0.0, epsilon=10e-8)
        model.compile(loss={'ctc': lambda y_true, y_pred: y_pred}, optimizer=opt)

        # captures output of softmax so we can decode the output during visualization
        test_func = K.function([input_data], [y_pred])

        default_logger.info('Create Model Successful, Compiles Model Successful. ')
        return model, model_data

    def ctc_lambda_func(self, args):
        y_pred, labels, input_length, label_length = args

        y_pred = y_pred[:, :, :]
        return K.ctc_batch_cost(labels, y_pred, input_length, label_length)

    def predict(self, data_input, input_len):
        """
        预测结果
        :param data_input:
        :param input_len:
        :return: 返回语音识别后的拼音符号列表
        """
        batch_size = 1
        in_len = np.zeros((batch_size), dtype=np.int32)
        in_len[0] = input_len
        x_in = np.zeros((batch_size, 1600, self.AUDIO_FEATURE_LENGTH, 1), dtype=np.float)

        for i in range(batch_size):
            x_in[i, 0:len(data_input)] = data_input

        base_pred = self.base_model.predict(x=x_in)
        base_pred = base_pred[:, :, :]
        r = K.ctc_decode(base_pred, in_len, greedy=True, beam_width=100, top_paths=1)
        r1 = K.get_value(r[0][0])
        # print('r1', r1)
        r1 = r1[0]
        return r1

    def recognize_speech(self, wavsignal, fs):
        """
        语音识别用的函数，识别一个wav序列的语音
        :param wavsignal:
        :param fs:
        :return:
        """
        self.check_initialized()
        data_input = get_frequency_features(wavsignal, fs)
        # t1=time.time()
        # print('time cost:',t1-t0)

        input_length = len(data_input)
        input_length = input_length // 8

        data_input = np.array(data_input, dtype=np.float)
        # print(data_input,data_input.shape)
        data_input = data_input.reshape(data_input.shape[0], data_input.shape[1], 1)
        # t2=time.time()
        r1 = self.predict(data_input, input_length)
        # t3=time.time()
        # print('time cost:',t3-t2)
        r_str = []
        for i in r1:
            r_str.append(self.pinyin_list[i])
        return r_str

    def recognize_speech_from_file(self, filename):
        """
        语音识别用的接口函数
        :param filename: 识别指定文件名的语音
        :return:
        """
        wavsignal, fs = read_wav_data(filename)
        return self.recognize_speech(wavsignal, fs)

    @property
    def model(self):
        """
        model
        :return: keras model
        """
        self.check_initialized()
        return self._model
